from __future__ import annotations

import itertools
import logging
import os
import re
from dataclasses import dataclass
from pathlib import Path
from typing import Match

from deptry.dependency import Dependency
from deptry.dependency_getter.base import DependenciesExtract, DependencyGetter


@dataclass
class RequirementsTxtDependencyGetter(DependencyGetter):
    """Extract dependencies from requirements.txt files."""

    requirements_txt: tuple[str, ...] = ("requirements.txt",)
    requirements_txt_dev: tuple[str, ...] = ("dev-requirements.txt", "requirements-dev.txt")

    def get(self) -> DependenciesExtract:
        dependencies = list(
            itertools.chain(
                *(self._get_dependencies_from_requirements_file(file_name) for file_name in self.requirements_txt)
            )
        )
        self._log_dependencies(dependencies=dependencies)

        dev_dependencies = list(
            itertools.chain(
                *(
                    self._get_dependencies_from_requirements_file(file_name)
                    for file_name in self._scan_for_dev_requirements_files()
                )
            )
        )
        self._log_dependencies(dependencies=dev_dependencies, is_dev=True)

        return DependenciesExtract(dependencies, dev_dependencies)

    def _scan_for_dev_requirements_files(self) -> list[str]:
        """
        Check if any of the files passed as requirements_txt_dev exist, and if so; return them.
        """
        dev_requirements_files = [file_name for file_name in self.requirements_txt_dev if file_name in os.listdir()]
        if dev_requirements_files:
            logging.debug("Found files with development requirements! %s", dev_requirements_files)
        return dev_requirements_files

    def _get_dependencies_from_requirements_file(self, file_name: str, is_dev: bool = False) -> list[Dependency]:
        logging.debug("Scanning %s for %s", file_name, "dev dependencies" if is_dev else "dependencies")
        dependencies = []

        file_path = Path(file_name)

        with file_path.open() as f:
            data = f.readlines()

        for line in data:
            dependency = self._extract_dependency_from_line(line, file_path)
            if dependency:
                dependencies.append(dependency)

        return dependencies

    def _extract_dependency_from_line(self, line: str, file_path: Path) -> Dependency | None:
        """
        Extract a dependency from a single line of a requirements.txt file.
        """
        line = self._remove_comments_from(line)
        line = self._remove_newlines_from(line)
        name = self._find_dependency_name_in(line)
        if name:
            line = line.replace(name, "")
            optional = self._check_if_dependency_is_optional(line)
            conditional = self._check_if_dependency_is_conditional(line)
            return Dependency(
                name=name,
                definition_file=file_path,
                optional=optional,
                conditional=conditional,
                module_names=self.package_module_name_map.get(name),
            )
        else:
            return None

    def _find_dependency_name_in(self, line: str) -> str | None:
        """
        Find the dependency name of a dependency specified according to the pip-standards for requirement.txt
        """
        if self._line_is_url(line):
            return self._extract_name_from_url(line)
        else:
            match = re.search("^[^-][a-zA-Z0-9-_]+", line)
            if match:
                return match.group(0)
        return None

    @staticmethod
    def _remove_comments_from(line: str) -> str:
        return re.sub(r"\s+#.*", "", line).strip()

    @staticmethod
    def _remove_newlines_from(line: str) -> str:
        return line.replace("\n", "")

    @staticmethod
    def _check_if_dependency_is_optional(line: str) -> bool:
        return bool(re.findall(r"\[([a-zA-Z0-9-]+?)\]", line))

    @staticmethod
    def _check_if_dependency_is_conditional(line: str) -> bool:
        return ";" in line

    @staticmethod
    def _line_is_url(line: str) -> Match[str] | None:
        return re.search(r"^(http|https|git\+https)", line)

    @staticmethod
    def _extract_name_from_url(line: str) -> str | None:
        # Try to find egg, for url like git+https://github.com/xxxxx/package@xxxxx#egg=package
        match = re.search("egg=([a-zA-Z0-9-_]*)", line)
        if match:
            return match.group(1)

        # for url like git+https://github.com/name/python-module.git@0d6dc38d58
        match = re.search(r"\/((?:(?!\/).)*?)\.git", line)
        if match:
            return match.group(1)

        # for url like https://github.com/urllib3/urllib3/archive/refs/tags/1.26.8.zip
        match = re.search(r"\/((?:(?!\/).)*?)\/archive\/", line)
        if match:
            return match.group(1)

        logging.warning("Could not parse dependency name from url %s", line)
        return None



from __future__ import annotations

import logging
from dataclasses import dataclass, field
from importlib.metadata import PackageNotFoundError, metadata
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from deptry.dependency import Dependency
    from deptry.imports.location import Location


@dataclass
class Module:
    """
    Represents an imported module and its properties.

    Attributes:
        name: The name of the imported module.
        standard_library: Whether the module is part of the Python standard library.
        local_module: Whether the module is a local module.
        package: The name of the package that contains the module.
        top_levels: A list of dependencies that contain this module in their top-level module
            names. This can be multiple, e.g. `google-cloud-api` and `google-cloud-bigquery` both have
            `google` in their top-level module names.
        dev_top_levels: A list of development dependencies that contain this module in their
            top-level module names. Can be multiple, similar to the attribute `top_levels`.
        is_provided_by_dependency: Whether the module is provided by a listed dependency.
        is_provided_by_dev_dependency: Whether the module is provided by a listed development dependency.
    """

    name: str
    standard_library: bool = False
    local_module: bool = False
    package: str | None = None
    top_levels: list[str] | None = None
    dev_top_levels: list[str] | None = None
    is_provided_by_dependency: bool | None = None
    is_provided_by_dev_dependency: bool | None = None

    def __post_init__(self) -> None:
        self._log()

    def _log(self) -> None:
        logging.debug("--- MODULE ---")
        logging.debug(self.__str__())
        logging.debug("")

    def __repr__(self) -> str:
        return f"Module '{self.name}'"

    def __str__(self) -> str:
        return "\n".join("{}: {}".format(*item) for item in vars(self).items())


@dataclass
class ModuleLocations:
    module: Module
    locations: list[Location] = field(default_factory=list)


class ModuleBuilder:
    def __init__(
        self,
        name: str,
        local_modules: set[str],
        stdlib_modules: frozenset[str],
        dependencies: list[Dependency] | None = None,
        dev_dependencies: list[Dependency] | None = None,
    ) -> None:
        """
        Create a Module object that represents an imported module.

        Args:
            name: The name of the imported module
            local_modules: The list of local modules
            stdlib_modules: The list of Python stdlib modules
            dependencies: A list of the project's dependencies
            dev_dependencies: A list of the project's development dependencies
        """
        self.name = name
        self.local_modules = local_modules
        self.stdlib_modules = stdlib_modules
        self.dependencies = dependencies or []
        self.dev_dependencies = dev_dependencies or []

    def build(self) -> Module:
        """
        Build the Module object. First check if the module is in the standard library or if it's a local module,
        in that case many checks can be omitted.
        """

        if self._in_standard_library():
            return Module(self.name, standard_library=True)

        if self._is_local_module():
            return Module(self.name, local_module=True)

        package = self._get_package_name_from_metadata()
        top_levels = self._get_corresponding_top_levels_from(self.dependencies)
        dev_top_levels = self._get_corresponding_top_levels_from(self.dev_dependencies)

        is_provided_by_dependency = self._has_matching_dependency(package, top_levels)
        is_provided_by_dev_dependency = self._has_matching_dev_dependency(package, dev_top_levels)
        return Module(
            self.name,
            package=package,
            top_levels=top_levels,
            dev_top_levels=dev_top_levels,
            is_provided_by_dependency=is_provided_by_dependency,
            is_provided_by_dev_dependency=is_provided_by_dev_dependency,
        )

    def _get_package_name_from_metadata(self) -> str | None:
        """
        Most packages simply have a field called "Name" in their metadata. This method extracts that field.
        """
        try:
            name: str = metadata(self.name)["Name"]
        except PackageNotFoundError:
            return None
        else:
            return name

    def _get_corresponding_top_levels_from(self, dependencies: list[Dependency]) -> list[str]:
        """
        Not all modules have associated metadata. e.g. `mpl_toolkits` from `matplotlib` has no metadata. However, it is
        in the top-level module names of package matplotlib. This function extracts all dependencies which have this
        module in their top-level module names.
        This can be multiple, e.g. `google-cloud-api` and `google-cloud-bigquery` both have `google` in their top-level
        module names.
        """
        return [
            dependency.name
            for dependency in dependencies
            if dependency.top_levels and self.name in dependency.top_levels
        ]

    def _in_standard_library(self) -> bool:
        return self.name in self.stdlib_modules

    def _is_local_module(self) -> bool:
        """
        Check if the module is a local directory with an __init__.py file.
        """
        return self.name in self.local_modules

    def _has_matching_dependency(self, package: str | None, top_levels: list[str]) -> bool:
        """
        Check if this module is provided by a listed dependency. This is the case if either the package name that was
        found in the metadata is listed as a dependency, or if we found a top-level module name match earlier.
        """
        return package and (package in [dep.name for dep in self.dependencies]) or len(top_levels) > 0

    def _has_matching_dev_dependency(self, package: str | None, dev_top_levels: list[str]) -> bool:
        """
        Same as _has_matching_dependency, but for development dependencies.
        """
        return package and (package in [dep.name for dep in self.dev_dependencies]) or len(dev_top_levels) > 0



from __future__ import annotations

import ast
from abc import ABC, abstractmethod
from collections import defaultdict
from dataclasses import dataclass
from typing import TYPE_CHECKING

import chardet

from deptry.imports.location import Location

if TYPE_CHECKING:
    from pathlib import Path


@dataclass
class ImportExtractor(ABC):
    """
    Base class for other classes that can be used to extract the imported modules from a file.
    """

    file: Path

    @abstractmethod
    def extract_imports(self) -> dict[str, list[Location]]:
        raise NotImplementedError()

    def _extract_imports_from_ast(self, tree: ast.AST) -> dict[str, list[Location]]:
        """
        Given an Abstract Syntax Tree, find the imported top-level modules.
        For example, given the source tree of a file with contents:

            from pandas.tools import scatter_matrix

        Will return the set {"pandas"}.
        """

        imported_modules: dict[str, list[Location]] = defaultdict(list)

        for node in ast.walk(tree):
            if isinstance(node, ast.Import):
                for module in node.names:
                    imported_modules[module.name.split(".")[0]].append(
                        Location(self.file, node.lineno, node.col_offset)
                    )
            elif isinstance(node, ast.ImportFrom) and node.module and node.level == 0:
                imported_modules[node.module.split(".")[0]].append(Location(self.file, node.lineno, node.col_offset))

        return imported_modules

    @staticmethod
    def _get_file_encoding(file: Path) -> str:
        with file.open("rb") as f:
            return chardet.detect(f.read())["encoding"]



from __future__ import annotations

import logging
import re
from contextlib import suppress
from importlib import metadata
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Sequence
    from importlib.metadata import Distribution
    from pathlib import Path


class Dependency:
    """
    A project's dependency with its associated top-level module names. There are stored in the metadata's top_level.txt.
    An example of this is 'matplotlib' with top-levels: ['matplotlib', 'mpl_toolkits', 'pylab'].

    Attributes:
        name (str): The name of the dependency.
        definition_file (Path): The path to the file defining the dependency, e.g. 'pyproject.toml'.
        is_conditional (bool): Indicates if the dependency is conditional.
        is_optional (bool): Indicates if the dependency is optional i.e. dependencies that are not installed by default
          and that can be used to create a variant of the package with a set of extra functionalities.
        found (bool): Indicates if the dependency has been found in the environment.
        top_levels (set[str]): The top-level module names associated with the dependency.
    """

    def __init__(
        self,
        name: str,
        definition_file: Path,
        conditional: bool = False,
        optional: bool = False,
        module_names: Sequence[str] | None = None,
    ) -> None:
        distribution = self.find_distribution(name)

        self.name = name
        self.definition_file = definition_file
        self.is_conditional = conditional
        self.is_optional = optional
        self.found = distribution is not None
        self.top_levels = self._get_top_levels(name, distribution, module_names)

    def _get_top_levels(
        self, name: str, distribution: Distribution | None, module_names: Sequence[str] | None
    ) -> set[str]:
        """
        Get the top-level module names for a dependency. They are searched for in the following order:
                1. If `module_names` is defined, simply use those as the top-level modules.
                2. Otherwise, try to extract the top-level module names from the metadata.
                3. If metadata extraction fails, fall back to the dependency's name, where `-` is replaced with `_`.

        Args:
            name: The name of the dependency.
            distribution: The metadata distribution of the package.
            module_names: If this is given, use these as the top-level modules instead of
                searching for them in the metadata.
        """
        if module_names is not None:
            return set(module_names)

        if distribution is not None:
            with suppress(FileNotFoundError):
                return self._get_top_level_module_names_from_top_level_txt(distribution)

            with suppress(FileNotFoundError):
                return self._get_top_level_module_names_from_record_file(distribution)

        # No metadata or other configuration has been found. As a fallback
        # we'll guess the name.
        module_name = name.replace("-", "_").lower()
        logging.warning(
            (
                "Assuming the corresponding module name of package %r is %r. Install the package or configure a"
                " package_module_name_map entry to override this behaviour."
            ),
            self.name,
            module_name,
        )
        return {module_name}

    def __repr__(self) -> str:
        return f"Dependency '{self.name}'"

    def __str__(self) -> str:
        return f"Dependency '{self.name}'{self._string_for_printing()}with top-levels: {self.top_levels}."

    @staticmethod
    def find_distribution(name: str) -> Distribution | None:
        try:
            return metadata.distribution(name)
        except metadata.PackageNotFoundError:
            return None

    def _string_for_printing(self) -> str:
        """
        Return 'Conditional', 'Optional' or 'Conditional and optional'
        """
        output_list = []
        if self.is_optional:
            output_list.append("optional")
        if self.is_conditional:
            output_list.append("conditional")

        if len(output_list) > 0:
            return f" ({', '.join(output_list)}) "
        else:
            return " "

    @staticmethod
    def _get_top_level_module_names_from_top_level_txt(distribution: Distribution) -> set[str]:
        """
        top-level.txt is a metadata file added by setuptools that looks as follows:

        610faff656c4cfcbb4a3__mypyc
        _black_version
        black
        blackd
        blib2to3

        This function extracts these names, if a top-level.txt file exists.
        """
        metadata_top_levels = distribution.read_text("top_level.txt")
        if metadata_top_levels is None:
            raise FileNotFoundError("top_level.txt")

        return {x for x in metadata_top_levels.splitlines() if x}

    @staticmethod
    def _get_top_level_module_names_from_record_file(distribution: Distribution) -> set[str]:
        """
        Get the top-level module names from the RECORD file, whose contents usually look as follows:

            ...
            ../../../bin/black,sha256=<HASH>,247
            __pycache__/_black_version.cpython-311.pyc,,
            _black_version.py,sha256=<HASH>,19
            black/trans.cpython-39-darwin.so,sha256=<HASH>
            black/trans.py,sha256=<HASH>
            blackd/__init__.py,sha256=<HASH>
            blackd/__main__.py,sha256=<HASH>
            ...

        So if no file top-level.txt is provided, we can try and extract top-levels from this file, in
        this case _black_version, black, and blackd.
        """
        metadata_records = distribution.read_text("RECORD")

        if metadata_records is None:
            raise FileNotFoundError("RECORD")

        matches = re.finditer(r"^(?!__)([a-zA-Z0-9-_]+)(?:/|\.py,)", metadata_records, re.MULTILINE)

        return {x.group(1) for x in matches}



from __future__ import annotations

import logging

SKIP_OBSOLETE_WARNING = (
    "Warning: In an upcoming release, support for the `--skip-obsolete` command-line option and the `skip_obsolete`"
    " configuration parameter will be discontinued. Instead, use the `--skip-unused` option or the `skip_unused`"
    " configuration parameter to achieve the desired behavior."
)
IGNORE_OBSOLETE_WARNING = (
    "Warning: In an upcoming release, support for the `--ignore-obsolete` command-line option and the `ignore_obsolete`"
    " configuration parameter will be discontinued. Instead, use the `--ignore-unused` option or the `ignore_unused`"
    " configuration parameter to achieve the desired behavior."
)
SKIP_OBSOLETE_AND_SKIP_UNUSED_ERROR_MESSAGE = (
    "Both `skip_obsolete` and `skip_unused` options are set, either in pyproject.toml or through their corresponding"
    " CLI arguments. Please use only the `skip_unused` option, as `skip_obsolete` will be deprecated in the future."
)
IGNORE_OBSOLETE_AND_IGNORE_UNUSED_ERROR_MESSAGE = (
    "Both `ignore_obsolete` and `ignore_unused` options are set, either in pyproject.toml or through their"
    " corresponding CLI arguments. Please use only the `ignore_unused` option, as `ignore_obsolete` will be deprecated"
    " in the future."
)


def get_value_for_skip_unused(skip_obsolete: bool, skip_unused: bool) -> bool:
    """
    The skip_obsolete argument will be deprecated in the future, users should use skip_unused instead.
    If only skip_obsolete is defined, display a warning message. If both skip_obsolete and skip_unused are defined,
    raise an Error.
    """
    if skip_obsolete:
        logging.warning(SKIP_OBSOLETE_WARNING)
        if skip_unused:
            raise ValueError(SKIP_OBSOLETE_AND_SKIP_UNUSED_ERROR_MESSAGE)
        else:
            return skip_obsolete
    return skip_unused


def get_value_for_ignore_unused(ignore_obsolete: tuple[str, ...], ignore_unused: tuple[str, ...]) -> tuple[str, ...]:
    """
    The ignore_obsolete argument will be deprecated in the future, users should use ignore_unused instead.
    If only ignore_obsolete is defined, display a warning message. If both ignore_obsolete and ignore_unused are defined,
    raise an Error.
    """
    if ignore_obsolete:
        logging.warning(IGNORE_OBSOLETE_WARNING)
        if ignore_unused:
            raise ValueError(IGNORE_OBSOLETE_AND_IGNORE_UNUSED_ERROR_MESSAGE)
        else:
            return ignore_obsolete
    return ignore_unused



from __future__ import annotations

import logging
from enum import Enum
from pathlib import Path

from deptry.exceptions import DependencySpecificationNotFoundError
from deptry.utils import load_pyproject_toml


class DependencyManagementFormat(Enum):
    PDM = "pdm"
    PEP_621 = "pep_621"
    POETRY = "poetry"
    REQUIREMENTS_TXT = "requirements_txt"


class DependencySpecificationDetector:
    """
    Class to detect how dependencies are specified:
    - Either find a pyproject.toml with a [poetry.tool.dependencies] section
    - Otherwise, find a pyproject.toml with a [tool.pdm] section
    - Otherwise, find a pyproject.toml with a [project] section
    - Otherwise, find a requirements.txt.

    """

    def __init__(self, config: Path, requirements_txt: tuple[str, ...] = ("requirements.txt",)) -> None:
        self.config = config
        self.requirements_txt = requirements_txt

    def detect(self) -> DependencyManagementFormat:
        pyproject_toml_found = self._project_contains_pyproject_toml()
        if pyproject_toml_found and self._project_uses_poetry():
            return DependencyManagementFormat.POETRY
        if pyproject_toml_found and self._project_uses_pdm():
            return DependencyManagementFormat.PDM
        if pyproject_toml_found and self._project_uses_pep_621():
            return DependencyManagementFormat.PEP_621
        if self._project_uses_requirements_txt():
            return DependencyManagementFormat.REQUIREMENTS_TXT

        raise DependencySpecificationNotFoundError(self.requirements_txt)

    def _project_contains_pyproject_toml(self) -> bool:
        if self.config.exists():
            logging.debug("pyproject.toml found!")
            return True
        else:
            logging.debug("No pyproject.toml found.")
            return False

    def _project_uses_poetry(self) -> bool:
        pyproject_toml = load_pyproject_toml(self.config)
        try:
            pyproject_toml["tool"]["poetry"]["dependencies"]
            logging.debug(
                "pyproject.toml contains a [tool.poetry.dependencies] section, so Poetry is used to specify the"
                " project's dependencies."
            )
        except KeyError:
            logging.debug(
                "pyproject.toml does not contain a [tool.poetry.dependencies] section, so Poetry is not used to specify"
                " the project's dependencies."
            )
            return False
        else:
            return True

    def _project_uses_pdm(self) -> bool:
        pyproject_toml = load_pyproject_toml(self.config)
        try:
            pyproject_toml["tool"]["pdm"]["dev-dependencies"]
            logging.debug(
                "pyproject.toml contains a [tool.pdm.dev-dependencies] section, so PDM is used to specify the project's"
                " dependencies."
            )
        except KeyError:
            logging.debug(
                "pyproject.toml does not contain a [tool.pdm.dev-dependencies] section, so PDM is not used to specify"
                " the project's dependencies."
            )
            return False
        else:
            return True

    def _project_uses_pep_621(self) -> bool:
        pyproject_toml = load_pyproject_toml(self.config)
        try:
            pyproject_toml["project"]
            logging.debug(
                "pyproject.toml contains a [project] section, so PEP 621 is used to specify the project's dependencies."
            )
        except KeyError:
            logging.debug(
                "pyproject.toml does not contain a [project] section, so PEP 621 is not used to specify the project's"
                " dependencies."
            )
            return False
        else:
            return True

    def _project_uses_requirements_txt(self) -> bool:
        check = any(Path(requirements_txt).is_file() for requirements_txt in self.requirements_txt)
        if check:
            logging.debug(
                "Dependency specification found in '%s'. Will use this to determine the project's dependencies.\n",
                ", ".join(self.requirements_txt),
            )
        return check



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pathlib import Path


@dataclass(frozen=True)
class Location:
    file: Path
    line: int | None = None
    column: int | None = None



from __future__ import annotations

import contextlib
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from deptry.dependency import Dependency
from deptry.dependency_getter.base import DependenciesExtract, DependencyGetter
from deptry.utils import load_pyproject_toml

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence


@dataclass
class PoetryDependencyGetter(DependencyGetter):
    """Extract Poetry dependencies from pyproject.toml."""

    def get(self) -> DependenciesExtract:
        dependencies = self._get_poetry_dependencies()
        self._log_dependencies(dependencies)

        dev_dependencies = self._get_poetry_dev_dependencies()
        self._log_dependencies(dev_dependencies, is_dev=True)

        return DependenciesExtract(dependencies, dev_dependencies)

    def _get_poetry_dependencies(self) -> list[Dependency]:
        pyproject_data = load_pyproject_toml(self.config)
        dependencies: dict[str, Any] = pyproject_data["tool"]["poetry"]["dependencies"]
        return self._get_dependencies(dependencies, self.package_module_name_map)

    def _get_poetry_dev_dependencies(self) -> list[Dependency]:
        """
        Poetry's development dependencies can be specified under either of the following:

        [tool.poetry.dev-dependencies]
        [tool.poetry.group.dev.dependencies]

        or both.
        """
        dependencies: dict[str, str] = {}
        pyproject_data = load_pyproject_toml(self.config)

        with contextlib.suppress(KeyError):
            dependencies = {**pyproject_data["tool"]["poetry"]["dev-dependencies"], **dependencies}

        with contextlib.suppress(KeyError):
            dependencies = {**pyproject_data["tool"]["poetry"]["group"]["dev"]["dependencies"], **dependencies}

        return self._get_dependencies(dependencies, self.package_module_name_map)

    def _get_dependencies(
        self, poetry_dependencies: dict[str, Any], package_module_name_map: Mapping[str, Sequence[str]]
    ) -> list[Dependency]:
        dependencies = []
        for dep, spec in poetry_dependencies.items():
            # dep is the dependency name, spec is the version specification, e.g. "^0.2.2" or {"*", optional = true}
            if dep != "python":
                optional = self._is_optional(spec)
                conditional = self._is_conditional(spec)
                dependencies.append(
                    Dependency(
                        dep,
                        self.config,
                        conditional=conditional,
                        optional=optional,
                        module_names=package_module_name_map.get(dep),
                    )
                )

        return dependencies

    @staticmethod
    def _is_optional(spec: str | dict[str, Any]) -> bool:
        """
        If a dependency specification is of the shape `isodate = {version = "*", optional = true}`, mark it as optional.
        """
        return bool(isinstance(spec, dict) and spec.get("optional"))

    @staticmethod
    def _is_conditional(spec: str | dict[str, Any]) -> bool:
        """
        If a dependency specification is of the shape `tomli = { version = "^2.0.1", python = "<3.11" }`, mark it as conditional.
        """
        return isinstance(spec, dict) and "python" in spec and "version" in spec



from __future__ import annotations

import logging
from collections import defaultdict
from typing import TYPE_CHECKING

from deptry.imports.extractors import NotebookImportExtractor, PythonImportExtractor

if TYPE_CHECKING:
    from pathlib import Path

    from deptry.imports.extractors.base import ImportExtractor
    from deptry.imports.location import Location


def get_imported_modules_for_list_of_files(list_of_files: list[Path]) -> dict[str, list[Location]]:
    logging.info("Scanning %d %s...", len(list_of_files), "files" if len(list_of_files) > 1 else "file")

    modules: dict[str, list[Location]] = defaultdict(list)

    for file in list_of_files:
        for module, locations in get_imported_modules_from_file(file).items():
            for location in locations:
                modules[module].append(location)

    logging.debug("All imported modules: %s\n", modules)

    return modules


def get_imported_modules_from_file(path_to_file: Path) -> dict[str, list[Location]]:
    logging.debug("Scanning %s...", path_to_file)

    modules = _get_extractor_class(path_to_file)(path_to_file).extract_imports()

    logging.debug("Found the following imports in %s: %s", path_to_file, modules)

    return modules


def _get_extractor_class(path_to_file: Path) -> type[ImportExtractor]:
    if path_to_file.suffix == ".ipynb":
        return NotebookImportExtractor
    return PythonImportExtractor



from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import TYPE_CHECKING

from deptry.issues_finder.base import IssuesFinder
from deptry.violations import TransitiveDependencyViolation

if TYPE_CHECKING:
    from deptry.module import Module
    from deptry.violations import Violation


@dataclass
class TransitiveDependenciesFinder(IssuesFinder):
    """
    Given a list of imported modules and a list of project dependencies, determine which ones are transitive.
    This is done by elimination; if a module uses an installed package but the package is;
    - not a local directory
    - not in standard library
    - not a dependency
    - not a dev dependency

    Then it must be a transitive dependency.
    """

    def find(self) -> list[Violation]:
        logging.debug("\nScanning for transitive dependencies...")
        transitive_dependencies: list[Violation] = []

        for module_with_locations in self.imported_modules_with_locations:
            module = module_with_locations.module

            logging.debug("Scanning module %s...", module.name)

            if self._is_transitive(module):
                # `self._is_transitive` only returns `True` if the package is not None.
                for location in module_with_locations.locations:
                    transitive_dependencies.append(TransitiveDependencyViolation(module, location))

        return transitive_dependencies

    def _is_transitive(self, module: Module) -> bool:
        if any(
            [
                module.package is None,
                module.is_provided_by_dependency,
                module.is_provided_by_dev_dependency,
                module.local_module,
            ]
        ):
            return False

        if module.name in self.ignored_modules:
            logging.debug("Dependency '%s' found to be a transitive dependency, but ignoring.", module.package)
            return False

        logging.debug("Dependency '%s' marked as a transitive dependency.", module.package)
        return True



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, ClassVar

from deptry.violations import Violation

if TYPE_CHECKING:
    from deptry.module import Module


@dataclass
class TransitiveDependencyViolation(Violation):
    error_code: ClassVar[str] = "DEP003"
    error_template: ClassVar[str] = "'{name}' imported but it is a transitive dependency"
    issue: Module

    def get_error_message(self) -> str:
        return self.error_template.format(name=self.issue.package)



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pathlib import Path


class DependencySpecificationNotFoundError(FileNotFoundError):
    def __init__(self, requirements_txt: tuple[str, ...]) -> None:
        super().__init__(
            "No file called 'pyproject.toml' with a [tool.poetry.dependencies], [tool.pdm] or [project] section or"
            f" file(s) called '{', '.join(requirements_txt)}' found. Exiting."
        )


class IncorrectDependencyFormatError(ValueError):
    def __init__(self) -> None:
        super().__init__("Incorrect dependency manage format. Only poetry, pdm and requirements.txt are supported.")


class PyprojectFileNotFoundError(FileNotFoundError):
    def __init__(self, directory: Path) -> None:
        super().__init__(f"No file `pyproject.toml` found in directory {directory}")


class UnsupportedPythonVersionError(ValueError):
    def __init__(self, version: tuple[int, int]) -> None:
        super().__init__(
            f"Python version {version[0]}.{version[1]} is not supported. Only versions >= 3.8 are supported."
        )



from __future__ import annotations

import ast
import itertools
import json
import logging
import re
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from deptry.imports.extractors.base import ImportExtractor

if TYPE_CHECKING:
    from pathlib import Path

    from deptry.imports.location import Location


@dataclass
class NotebookImportExtractor(ImportExtractor):
    """Extract import statements from a Jupyter notebook."""

    def extract_imports(self) -> dict[str, list[Location]]:
        """Extract the imported top-level modules from all code cells in the Jupyter Notebook."""
        notebook = self._read_ipynb_file(self.file)
        if not notebook:
            return {}

        cells = self._keep_code_cells(notebook)
        import_statements = [self._extract_import_statements_from_cell(cell) for cell in cells]
        tree = ast.parse("\n".join(itertools.chain.from_iterable(import_statements)), str(self.file))
        return self._extract_imports_from_ast(tree)

    @classmethod
    def _read_ipynb_file(cls, path_to_ipynb: Path) -> dict[str, Any] | None:
        try:
            with path_to_ipynb.open() as ipynb_file:
                notebook: dict[str, Any] = json.load(ipynb_file)
        except (UnicodeDecodeError, ValueError):
            try:
                with path_to_ipynb.open(encoding=cls._get_file_encoding(path_to_ipynb)) as ipynb_file:
                    notebook = json.load(ipynb_file, strict=False)
            except UnicodeDecodeError:
                logging.warning("Warning: File %s could not be decoded. Skipping...", path_to_ipynb)
                return None
        return notebook

    @staticmethod
    def _keep_code_cells(notebook: dict[str, Any]) -> list[dict[str, Any]]:
        return [cell for cell in notebook["cells"] if cell["cell_type"] == "code"]

    @staticmethod
    def _contains_import_statements(line: str) -> bool:
        return re.search(r"^(?:from\s+(\w+)(?:\.\w+)?\s+)?import\s+([^\s,.]+)(?:\.\w+)?", line) is not None

    @classmethod
    def _extract_import_statements_from_cell(cls, cell: dict[str, Any]) -> list[str]:
        return [line for line in cell["source"] if cls._contains_import_statements(line)]



from __future__ import annotations

import ast
import logging
from dataclasses import dataclass
from typing import TYPE_CHECKING

from deptry.imports.extractors.base import ImportExtractor

if TYPE_CHECKING:
    from deptry.imports.location import Location


@dataclass
class PythonImportExtractor(ImportExtractor):
    """Extract import statements from a Python module."""

    def extract_imports(self) -> dict[str, list[Location]]:
        """Extract all imported top-level modules from the Python file."""
        try:
            with self.file.open() as python_file:
                tree = ast.parse(python_file.read(), str(self.file))
        except (UnicodeDecodeError, ValueError):
            try:
                with self.file.open(encoding=self._get_file_encoding(self.file)) as python_file:
                    tree = ast.parse(python_file.read(), str(self.file))
            except UnicodeDecodeError:
                logging.warning("Warning: File %s could not be decoded. Skipping...", self.file)
                return {}

        return self._extract_imports_from_ast(tree)



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING, ClassVar

if TYPE_CHECKING:
    from deptry.dependency import Dependency
    from deptry.imports.location import Location
    from deptry.module import Module


@dataclass
class Violation(ABC):
    """
    An abstract base class representing a violation found in the project's dependencies.

    Attributes:
        error_code: A class variable representing the error code associated with the violation. e.g.
            `DEP001`.
        error_template: A class variable representing a string template used to generate an error
            message for the violation.
        issue: An attribute representing the module or dependency where the violation
            occurred.
        location: An attribute representing the location in the code where the violation occurred.
    """

    error_code: ClassVar[str] = ""
    error_template: ClassVar[str] = ""
    issue: Dependency | Module
    location: Location

    @abstractmethod
    def get_error_message(self) -> str:
        raise NotImplementedError()



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from deptry.utils import load_pyproject_toml

if TYPE_CHECKING:
    from pathlib import Path

    import click


def read_configuration_from_pyproject_toml(ctx: click.Context, _param: click.Parameter, value: Path) -> Path | None:
    """
    Callback that, given a click context, overrides the default values with configuration options set in a
    pyproject.toml file.
    Using a callback ensures that the following order is respected for setting an option:
    1. Default value is set
    2. Value is overrode by the one set from pyproject.toml, if any
    3. Value is overrode by the one set from the command line, if any
    """

    try:
        pyproject_data = load_pyproject_toml(value)
    except FileNotFoundError:
        logging.debug("No pyproject.toml file to read configuration from.")
        return value

    try:
        deptry_toml_config = pyproject_data["tool"]["deptry"]
    except KeyError:
        logging.debug("No configuration for deptry was found in pyproject.toml.")
        return value

    click_default_map: dict[str, Any] = {}

    if ctx.default_map:
        click_default_map.update(ctx.default_map)

    click_default_map.update(deptry_toml_config)

    ctx.default_map = click_default_map

    return value



from __future__ import annotations

import logging
import os
import re
from dataclasses import dataclass
from pathlib import Path
from typing import Pattern

from pathspec import PathSpec


@dataclass
class PythonFileFinder:
    """
    Get a list of all .py and .ipynb files recursively within a directory.
    Args:
        exclude: A list of regex patterns of paths to ignore.
        extend_exclude: An additional list of regex patterns of paths to ignore.
        using_default_exclude: Whether the exclude list was explicitly set, or the default was used.
        ignore_notebooks: If ignore_notebooks is set to True, .ipynb files are ignored and only .py files are returned.
    """

    exclude: tuple[str, ...]
    extend_exclude: tuple[str, ...]
    using_default_exclude: bool
    ignore_notebooks: bool = False

    def get_all_python_files_in(self, directories: tuple[Path, ...]) -> list[Path]:
        logging.debug("Collecting Python files to scan...")

        source_files = set()

        ignore_regex = re.compile("|".join(self.exclude + self.extend_exclude))
        file_lookup_suffixes = {".py"} if self.ignore_notebooks else {".py", ".ipynb"}

        gitignore_spec = self._generate_gitignore_pathspec(Path("."))

        for directory in directories:
            for root_str, dirs, files in os.walk(directory, topdown=True):
                root = Path(root_str)

                if self._is_directory_ignored(root, ignore_regex):
                    dirs[:] = []
                    continue

                for file_str in files:
                    file = root / file_str
                    if not self._is_file_ignored(file, file_lookup_suffixes, ignore_regex, gitignore_spec):
                        source_files.add(file)

        source_files_list = list(source_files)

        logging.debug("Python files to scan for imports:\n%s\n", "\n".join([str(file) for file in source_files_list]))

        return source_files_list

    def _is_directory_ignored(self, directory: Path, ignore_regex: Pattern[str]) -> bool:
        return bool((self.exclude + self.extend_exclude) and ignore_regex.match(str(directory)))

    def _is_file_ignored(
        self, file: Path, file_lookup_suffixes: set[str], ignore_regex: Pattern[str], gitignore_spec: PathSpec | None
    ) -> bool:
        return bool(
            file.suffix not in file_lookup_suffixes
            or ((self.exclude + self.extend_exclude) and ignore_regex.match(file.as_posix()))
            or (gitignore_spec and gitignore_spec.match_file(file))
        )

    def _generate_gitignore_pathspec(self, directory: Path) -> PathSpec | None:
        # If `exclude` is explicitly set, `.gitignore` is not taken into account.
        if not self.using_default_exclude:
            return None

        try:
            with (directory / ".gitignore").open() as gitignore:
                return PathSpec.from_lines("gitwildmatch", gitignore)
        except FileNotFoundError:
            return None



from __future__ import annotations

from deptry.imports.extractors.notebook_import_extractor import NotebookImportExtractor
from deptry.imports.extractors.python_import_extractor import PythonImportExtractor

__all__ = (
    "NotebookImportExtractor",
    "PythonImportExtractor",
)



from __future__ import annotations

from deptry.violations.base import Violation
from deptry.violations.misplaced_dev import MisplacedDevDependencyViolation
from deptry.violations.missing import MissingDependencyViolation
from deptry.violations.transitive import TransitiveDependencyViolation
from deptry.violations.unused import UnusedDependencyViolation

__all__ = (
    "MisplacedDevDependencyViolation",
    "MissingDependencyViolation",
    "UnusedDependencyViolation",
    "TransitiveDependencyViolation",
    "Violation",
)
